﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.CombinedIntegerVectorEncoding {
  [StorableClass]
  [Item("CombinedIntegerVector", "Represents a combined vector of integer values.")]
  public class CombinedIntegerVector : IntegerVector, IMatching {

    [Storable]
    protected int actionPart;

    /// <summary>
    /// The bounds must contain at least one row with two columns. The first two columns specify min and max values.
    /// The max value -1 in a row is the don't care symbol for LCS.
    /// Each row represents the bounds for a certain dimension. If fewer rows are given, the rows are cycled
    /// </summary>
    [Storable]
    protected int[,] bounds;

    [StorableConstructor]
    protected CombinedIntegerVector(bool deserializing) : base(deserializing) { }
    public CombinedIntegerVector() : base() { }

    public CombinedIntegerVector(int[] elements, int actionPart, int[,] bounds)
      : base(elements) {
      this.actionPart = actionPart;
      //check if combinedVector satisfies bounds and if bounds is set correctly (at leats one row with 2 columns) is missing!
      this.bounds = bounds;
    }

    public CombinedIntegerVector(IntegerVector combinedVector, int actionPart, int[,] bounds)
      : this(combinedVector.ToArray(), actionPart, bounds) { }

    public CombinedIntegerVector(IntegerVector condition, IntegerVector action, int[,] bounds)
      : base(condition.Concat(action).ToArray()) {
      actionPart = action.Length;
      //check if combinedVector satisfies bounds and if bounds is set correctly (at leats one row with 2 columns) is missing!
      this.bounds = bounds;
    }

    public CombinedIntegerVector(IRandom random, int length, int min, int max, int actionLenght, int actionMin, int actionMax) :
      this(new IntegerVector(length, random, min, max), new IntegerVector(actionLenght, random, actionMin, actionMax), null) {
      bounds = new int[length + actionLenght, 2];
      for (int i = 0; i < length; i++) {
        bounds[i, 0] = min;
        bounds[i, 1] = max;
      }
      for (int i = length; i < length + actionLenght; i++) {
        bounds[i, 0] = actionMin;
        bounds[i, 1] = actionMax;
      }
    }

    public CombinedIntegerVector(int length, int actionPart, int[,] bounds)
      : base(length) {
      this.actionPart = actionPart;
      this.bounds = bounds;
    }

    protected CombinedIntegerVector(CombinedIntegerVector original, Cloner cloner)
      : base(original, cloner) {
      actionPart = original.actionPart;
      bounds = (int[,])original.bounds.Clone();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new CombinedIntegerVector(this, cloner);
    }

    public IMatching Condition {
      get {
        int[] condition = new int[Length - actionPart];
        Array.Copy(this.array, condition, Length - actionPart);
        return new CombinedIntegerVector(condition, 0, bounds);
      }
    }

    public IMatching Action {
      get {
        int[] action = new int[actionPart];
        Array.Copy(this.array, Length - actionPart, action, 0, actionPart);
        return new CombinedIntegerVector(action, actionPart, bounds);
      }
    }

    public bool MatchCondition(IMatching target) {
      var targetVector = target as CombinedIntegerVector;
      if (targetVector == null) return false;
      if (targetVector.Length != this.Length) return false;

      int curbounds;
      for (int i = 0; i < this.Length - this.actionPart; i++) {
        curbounds = i % bounds.GetLength(0);
        //if don't care symbol is matched, next indices can be checked
        if (this[i].Equals(bounds[curbounds, 1] - 1)) {
          continue;
        }
        if (!this[i].Equals(targetVector[i])) {
          return false;
        }
      }
      return true;
    }

    // no "don't care" symbols have to be considered
    public bool MatchAction(IMatching target) {
      var targetVector = target as CombinedIntegerVector;
      if (targetVector == null) return false;
      if (targetVector.actionPart != this.actionPart) return false;

      int curPos = this.Length - this.actionPart;
      int curTargetPos = targetVector.Length - targetVector.actionPart;
      for (int i = 0; i < this.actionPart; i++) {
        if (!this[curPos + i].Equals(targetVector[curTargetPos + i])) {
          return false;
        }
      }
      return true;
    }

    public override string ToString() {
      StringBuilder strBuilder = new StringBuilder(this.Length);
      strBuilder.Append('[');
      int curbounds;
      for (int i = 0; i < this.Length; i++) {
        curbounds = i % bounds.GetLength(0);
        if (i >= this.Length - this.actionPart || this[i] < bounds[curbounds, 1] - 1) {
          strBuilder.Append(this[i]);
        } else {
          strBuilder.Append('#');
        }
        if (i < this.Length - 1) {
          strBuilder.Append(';');
        }
      }
      strBuilder.Append(']');
      return strBuilder.ToString();
    }

    public override bool Equals(object obj) {
      var cast = obj as CombinedIntegerVector;
      if (cast != null) {
        return this.Equals(cast);
      }
      return false;
    }

    public bool Equals(IMatching other) {
      return this.MatchAction(other) && this.MatchCondition(other);
    }

    public override int GetHashCode() {
      int result = actionPart;
      for (int i = 0; i < array.Length; i++) {
        result ^= array[i];
      }
      return result.GetHashCode();
    }
  }
}
